import click
from .PushshiftAPI import PushshiftAPI
from . import writers as wt
from . import utilities as ut
from pathlib import Path
import pprint


@click.command(context_settings=dict(max_content_width=100))
@click.argument('search_type', type=click.Choice(['comments', 'submissions']), default='comments')
@click.option("-q", "--query", help='search term(s)', type=str)
@click.option("-s", "--subreddits", help='restrict search to subreddit(s)', type=str)
@click.option("-a", "--authors", help='restrict search to author(s)', type=str)
@click.option("-l", "--limit", default=20, help='maximum number of items to retrieve')
@click.option("--before", help='restrict to results before date '
                               '(datetime or int + s,m,h,d; eg, 30d for 30 days)', type=str)
@click.option("--after", help='restrict to results after date '
                              '(datetime or int + s,m,h,d; eg, 30d for 30 days)', type=str)
@click.option("-o", "--output", type=click.Path(),
              help="output file for saving all results in a single file")
@click.option("--output-template", type=str,
              help="""
              output file name template for saving each result in a separate file
              template can include output directory and fields from each result
              note, if using --filter, any fields in output-template MUST be
              included in filtered fields
              
              example:
              'output_path/{author}.{id}.csv'
              'output_path/{subreddit}_{created_utc}.json'

              """)
@click.option('--format', type=click.Choice(['json', 'csv']), default='csv')
@click.option("-f", "--filter", "filter_", type=str,
              help="filter fields to retrieve (must be in quotes or have no spaces), defaults to all")
@click.option("--prettify", is_flag=True, default=False,
              help="make output slightly less ugly (for json only)")
@click.option("--dry-run", is_flag=True, default=False,
              help="print potential names of output files, but don't actually write any files")
@click.option("--no-output-template-check", is_flag=True, default=False)
@click.option("--proxy", type=str, default=None)
@click.option("--verbose", is_flag=True, default=False)
def cli(search_type, query, subreddits, authors, limit, before, after,
        output, output_template, format, filter_, prettify, dry_run,
        no_output_template_check, proxy, verbose):
    """
    retrieve comments or submissions from reddit which meet given criteria

    """

    if output is None and output_template is None:
        raise click.UsageError("must supply either --output or --output-template")

    if output is not None and output_template is not None:
        raise click.UsageError("can only supply --output or --output-template, not both")

    verbose = verbose or dry_run

    if output:
        batch_mode = True
    else:
        batch_mode = False

    api = PushshiftAPI(https_proxy=proxy)
    search_args = dict()

    query = ut.string_to_list(query)
    filter_ = ut.string_to_list(filter_)
    authors = ut.string_to_list(authors)
    subreddits = ut.string_to_list(subreddits)
    before = ut.string_to_epoch(before)
    after = ut.string_to_epoch(after)

    # use a dict to pass args to search function because certain parameters
    # don't have defaults (eg, passing filter=None returns no fields)
    search_args = ut.build_search_kwargs(
        search_args,
        q=query,
        subreddit=subreddits,
        author=authors,
        limit=limit,
        before=before,
        after=after,
        filter=filter_,
    )

    search_functions = {
        'comments': api.search_comments,
        'submissions': api.search_submissions,
    }[search_type]

    if verbose:
        click.echo("calling api with following arguments:")
        click.echo(pprint.pformat(search_args))

    things = search_functions(**search_args)
    thing, things = ut.peek_first_item(things)
    if thing is None:
        click.secho("no results found", err=True, bold=True)
        return

    fields, missing_fields = ut.validate_fields(thing, filter_)

    if missing_fields:
        missing_fields = sorted(missing_fields)
        click.secho("server did not return following fields: {}".format(missing_fields),
                    bold=True, err=True)

    writer_class = choose_writer_class(format, batch_mode)
    writer = writer_class(fields=fields, prettify=prettify)

    if batch_mode:
        save_to_single_file(things, output, writer=writer,
                            count=limit, verbose=verbose, dry_run=dry_run)
    else:
        if not no_output_template_check:
            validate_output_template(output_template)

        save_to_multiple_files(things, output_template, writer=writer,
                               count=limit, verbose=verbose, dry_run=dry_run)


def choose_writer_class(format, batch_mode):
    """
    Choose appropriate writer class

    :param format: str
    :param batch_mode: bool
    :return: Class

    """
    writer_cls = {
        ('json', False): wt.JsonWriter,
        ('json', True): wt.JsonBatchWriter,
        ('csv', False): wt.CsvWriter,
        ('csv', True): wt.CsvBatchWriter,
    }[(format, batch_mode)]

    return writer_cls


def save_to_single_file(things, output_file, writer, count,
                        verbose=False, dry_run=False):
    """
    Write all things to a single file

    :param things: iterable
    :param output_file: str
    :param writer: Writer
    :param count: int
    :param verbose: bool
    :param dry_run: bool

    """
    count = 0
    writer.open(output_file)
    writer.header()
    try:
        with click.progressbar(things, length=count) as things:
            for thing in things:
                if not dry_run:
                    writer.write(thing.d_)
                    count += 1
    finally:
        writer.footer()
        writer.close()

    if verbose:
        click.echo("wrote {} items to {}".format(count, output_file))


def save_to_multiple_files(things, output_template, writer, count,
                           verbose=False, dry_run=False):
    """
    Write things to a separate file per thing

    :param things: iterable
    :param output_template: str
        template should contain python str format codes, such as "{subreddit}.{id}.json"
    :param writer: Writer
    :param count: int
    :param verbose: bool
    :param dry_run: bool

    """
    if dry_run:
        progressbar = ut.DummyProgressBar(things)
    else:
        progressbar = click.progressbar(things, length=count)

    count = 0
    with progressbar as things:
        for thing in things:
            output_file = output_template.format(**thing.d_)
            p = Path(output_file)
            if not p.parent.exists():
                p.parent.mkdir(parents=True)

            if dry_run:
                click.echo("saving to: {}".format(output_file))
            else:
                try:
                    writer.open(output_file)
                    writer.header()
                    writer.write(thing.d_)
                    writer.footer()
                    count += 1
                finally:
                    writer.close()

    if verbose:
        click.echo("wrote {} items".format(count))


def validate_output_template(output_template):
    """
    Crude sanity check that output template looks reasonable

    :param output_template: str

    """
    if "{" not in output_template or "}" not in output_template:
        raise click.BadParameter("output_template must include python "
                                 "string formatting replacement fields\n"
                                 "use --no-output-template-check to override check")

    try:
        p = Path(output_template)
    except:
        raise click.BadParameter("output_template does not look like "
                                 "a valid path\n"
                                 "use --no-output-template-check to override check")

    if len(p.parts) > 4:
        raise click.BadParameter("output_template looks like it is going to "
                                 "create a lot of directories\n"
                                 "use --no-output-template-check to override check")


if __name__ == '__main__':
    cli()






from collections import namedtuple, deque, Counter
import copy
import json
import requests
import time
from datetime import datetime as dt
import warnings

class RateLimitCache(object):
    def __init__(self, n, t=60):
        self.n = n
        self.t = t
        self.cache = deque()
    @property
    def delta(self):
        """Time since earliest call"""
        if len(self.cache) == 0:
            return 0
        return (time.time() - self.cache[0])
    def update(self):
        while self.delta > self.t:
            try:
                self.cache.popleft()
            except IndexError:
                return
    @property
    def blocked(self):
        """Test if additional calls need to be blocked"""
        self.update()
        return len(self.cache) >= self.n
    @property
    def interval(self):
        self.update()
        if self.t > self.delta:
             return self.t - self.delta
        else:
            return 0
    def new(self):
        self.update()
        if self.blocked:
            raise Exception("RateLimitCache is blocked.")
        self.cache.append(time.time())


class PushshiftAPIMinimal(object):
    #base_url = {'search':'https://api.pushshift.io/reddit/{}/search/',
    #            'meta':'https://api.pushshift.io/meta/'}
    _base_url = 'https://{domain}.pushshift.io/{{endpoint}}'
    _limited_args = ('aggs', 'ids')
    _thing_prefix = {
        	'Comment':'t1_',
        	'Account':'t2_',
        	'Submission':'t3_',
        	'Message':'t4_',
        	'Subreddit':'t5_',
        	'Award':'t6_'
    }
    def __init__(self,
                 max_retries=20,
                 max_sleep=3600,
                 backoff=2,
                 rate_limit_per_minute=None,
                 max_results_per_request=500,
                 detect_local_tz=True,
                 utc_offset_secs=None,
                 domain='api',
                 https_proxy=None
                ):
        assert max_results_per_request <= 500
        assert backoff >= 1

        self.max_retries = max_retries
        self.max_sleep   = max_sleep
        self.backoff     = backoff
        self.max_results_per_request = max_results_per_request

        self._utc_offset_secs = utc_offset_secs
        self._detect_local_tz = detect_local_tz

        self.domain = domain
        if https_proxy is not None:
            self.proxies = {"https": https_proxy }
        else:
            self.proxies = {}

        if rate_limit_per_minute is None:
            response = self._get(self.base_url.format(endpoint='meta'))
            rate_limit_per_minute = response['server_ratelimit_per_minute']
        self._rlcache = RateLimitCache(n=rate_limit_per_minute, t=60)

    @property
    def base_url(self):
        return self._base_url.format(domain=self.domain)

    @property
    def utc_offset_secs(self):
        if self._utc_offset_secs is not None:
            return self._utc_offset_secs

        if self._detect_local_tz:
            try:
                self._utc_offset_secs = dt.utcnow().astimezone().utcoffset().total_seconds()
            except ValueError:
                self._utc_offset_secs = 0
        else:
            self._utc_offset_secs = 0

        return self._utc_offset_secs

    def _limited(self, payload):
        """Turn off bells and whistles for special API endpoints"""
        return any(arg in payload for arg in self._limited_args)

    def _epoch_utc_to_local(self, epoch):
        return epoch - self.utc_offset_secs

    def _wrap_thing(self, thing, kind):
        """Mimic praw.Submission and praw.Comment API"""
        thing['created'] = self._epoch_utc_to_local(thing['created_utc'])
        thing['d_'] = copy.deepcopy(thing)
        ThingType = namedtuple(kind, thing.keys())
        thing = ThingType(**thing)
        return thing

    def _impose_rate_limit(self, nth_request=0):
        interval = 0
        if  hasattr(self, '_rlcache'):
            if self._rlcache.blocked:
                interval = self._rlcache.interval
        interval = max(interval, self.backoff*nth_request)
        interval = min(interval, self.max_sleep)
        if interval > 0:
            time.sleep(interval)

    def _add_nec_args(self, payload):
        """Adds 'limit' and 'created_utc' arguments to the payload as necessary."""
        if self._limited(payload):
            # Do nothing I guess? Not sure how paging works on this endpoint...
            return
        if 'limit' not in payload:
            payload['limit'] = self.max_results_per_request
        if 'sort' not in payload:
            # Getting weird results if this is not made explicit. Unclear why.
            payload['sort'] = 'desc'
        if 'filter' in payload: #and payload.get('created_utc', None) is None:
            if not isinstance(payload['filter'], list):
                if isinstance(payload['filter'], str):
                    payload['filter'] = [payload['filter']]
                else:
                    payload['filter'] = list(payload['filter'])
            if 'created_utc' not in payload['filter']:
                payload['filter'].append('created_utc')

    def _get(self, url, payload={}):
        i, success = 0, False
        while (not success) and (i<self.max_retries):
            if i > 0:
                warnings.warn("Unable to connect to pushshift.io. Retrying after backoff.")
            self._impose_rate_limit(i)
            i+=1
            try:
                response = requests.get(url, params=payload, proxies=self.proxies)
            except requests.ConnectionError:
                continue
            success = response.status_code == 200
            if not success:
                warnings.warn("Got non 200 code %s" % response.status_code)
        if not success:
            raise Exception("Unable to connect to pushshift.io. Max retries exceeded.")
        return json.loads(response.text)

    def _handle_paging(self, url):
        limit = self.payload.get('limit', None)
        #n = 0
        while True:
            if limit is not None:
                if limit > self.max_results_per_request:
                    self.payload['limit'] = self.max_results_per_request
                    limit -= self.max_results_per_request
                else:
                    self.payload['limit'] = limit
                    limit = 0
            elif 'ids' in self.payload:
                limit = 0
                if len(self.payload['ids']) > self.max_results_per_request:
                    err_msg = "When searching by ID, number of IDs must be fewer than the max number of objects in a single request ({})."
                    raise NotImplementedError(err_msg.format(self.max_results_per_request))
            self._add_nec_args(self.payload)

            yield self._get(url, self.payload)

            if (limit is not None) & (limit == 0):
                return

    def _search(self,
                kind,
                stop_condition=lambda x: False,
                return_batch=False,
                dataset='reddit',
                **kwargs):
        self.payload = copy.deepcopy(kwargs)
        endpoint = '{dataset}/{kind}/search'.format(dataset=dataset, kind=kind)
        url = self.base_url.format(endpoint=endpoint)
        for response in self._handle_paging(url):
            if 'aggs' in response:
                yield response['aggs']
                # Aggs responses are unreliable in subsequent batches with
                # current search paging implementation. Enforce aggs result
                # is only returned once.
                self.payload.pop('aggs')
            results = response['data']
            if len(results) == 0:
                return
            if return_batch:
                batch = []

            for thing in results:
                thing = self._wrap_thing(thing, kind)

                if return_batch:
                    batch.append(thing)
                else:
                    yield thing

                if stop_condition(thing):
                    if return_batch:
                        return batch
                    return

            if return_batch:
                yield batch

            # For paging.
            if self.payload.get('sort') == 'desc':
                self.payload['before'] = thing.created_utc
            else:
                self.payload['after'] = thing.created_utc


#class PushshiftAPI(PushshiftAPIMinimal):
    # Fill out this class with more user-friendly features later
#    pass

class PushshiftAPI(PushshiftAPIMinimal):
    def __init__(self, r=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.r = r
        self._search_func = self._search
        if r is not None:
            self._search_func = self._praw_search

    def search_comments(self, **kwargs):
        return self._search_func(kind='comment', **kwargs)

    def search_submissions(self, **kwargs):
        return self._search_func(kind='submission', **kwargs)

    def redditor_subreddit_activity(self, author, **kwargs):
        outv = {}
        for k in ('comment', 'submission'):
            agg = next(self._search(kind=k, author=author, aggs='subreddit', **kwargs))
            outv[k] = Counter({rec['key']:rec['doc_count'] for rec in agg['subreddit']})
        return outv

    def _get_submission_comment_ids(self, submission_id, **kwargs):
        self.payload = copy.deepcopy(kwargs)
        endpoint = 'reddit/submission/comment_ids/{}'.format(submission_id)
        url = self.base_url.format(endpoint=endpoint)
        return self._get(url, self.payload)['data']

    def _praw_search(self, **kwargs):
        prefix = self._thing_prefix[kwargs['kind'].title()]

        self.payload = copy.deepcopy(kwargs)

        client_return_batch = kwargs.get('return_batch')
        if client_return_batch is False:
            self.payload.pop('return_batch')

        if 'filter' in kwargs:
            self.payload.pop('filter')


        gen = self._search(return_batch=True, filter='id', **self.payload)
        using_gsci = False
        if kwargs.get('kind') == 'comment' and self.payload.get('submission_id'):
            using_gsci = True
            gen = [self._get_submission_comment_ids(**kwargs)]

        for batch in gen:
            if not batch:
                return
            if using_gsci:
                fullnames = [prefix + base36id for base36id in batch]
            else:
                fullnames = [prefix + c.id for c in batch]
            praw_batch = self.r.info(fullnames=fullnames)
            if client_return_batch:
                yield praw_batch
            else:
                for praw_thing in praw_batch:
                    yield praw_thing



import json
import csv

from .utilities import slice_dict


class Writer(object):
    """
    Base Writer class

    """
    def __init__(self, fields):
        self.fields = fields
        self.fp = None

    def header(self):
        """
        Write header to output file if necessary

        """
        pass

    def footer(self):
        """
        Write footer to output file if necessary

        """
        pass

    def open(self, fp):
        """
        Open output file for writing if necessary

        """
        if hasattr(fp, 'write'):
            self.fp = fp
        else:
            self.fp = open(fp, 'w', encoding='utf8', newline='')

    def close(self):
        """
        Close output file if necessary

        """
        if hasattr(self.fp, 'close_intelligently'):
            self.fp.close_intelligently()
        else:
            self.fp.close()


class JsonWriter(Writer):
    """
    Output comments/submissions in JSON format, all things to a single file

    """
    def __init__(self, fields, prettify=False, delimiter=',', **kwargs):
        super().__init__(fields=fields)
        self.prettify = prettify
        self.delimiter = delimiter
        self.items = 0

        if prettify:
            self.indent = 2
            self.delimiter = self.delimiter + '\n'
        else:
            self.indent = None

    def write(self, obj):
        obj = slice_dict(obj, self.fields)
        json.dump(obj, self.fp, indent=self.indent)
        self.items += 1


class JsonBatchWriter(JsonWriter):
    """
    Output comments/submissions in JSON format, one file per thing

    """

    def header(self):
        self.fp.write('[')

    def footer(self):
        self.fp.write(']')

    def write(self, obj):
        obj = slice_dict(obj, self.fields)

        if self.items > 0:
            # we've already written something, so
            # append a comma to make this a json list
            self.fp.write(self.delimiter)

        json.dump(obj, self.fp, indent=self.indent)
        self.items += 1


class CsvWriter(Writer):
    """
    Output comments/submissions in CSV format, one file per thing

    """
    def __init__(self, fields, delimiter=',', **kwargs):
        super().__init__(fields=fields)
        self.items = 0
        self.writer = None
        self.delimiter = delimiter

    def open(self, fp):
        super().open(fp)
        self.writer = csv.DictWriter(self.fp,
                                     delimiter=self.delimiter,
                                     fieldnames=self.fields)

    def header(self):
        self.writer.writeheader()

    def write(self, obj):
        obj = slice_dict(obj, self.fields)
        self.writer.writerow(obj)
        self.items += 1


class CsvBatchWriter(CsvWriter):
    """
    Output comments/submissions in CSV format, all to a single file
    """
    # defined just for consistency with Json/JsonBatch
    pass



import itertools
import dateutil.parser as dp
import re
import click


def validate_fields(item, fields=None):
    """
    Check that all requested fields were returned

    :param item: comment or submission
    :param fields: list[str]
    :return: list[str]
    """
    actual_fields = item.d_.keys()

    if fields is None:
        requested_fields = actual_fields
    else:
        requested_fields = fields

    missing_fields = set(requested_fields).difference(actual_fields)

    # drop extra fields returned from api
    final_fields = set(requested_fields).intersection(actual_fields)
    return final_fields, missing_fields


def peek_first_item(gen):
    """
    Peek at first item from generator if available, else return None

    :param gen: generator
    :return: first item, generator

    """
    try:
        item = next(gen)
    except StopIteration:
        item = None

    gen = itertools.chain([item], gen)

    return item, gen


def slice_dict(d, keys):
    """
    Only return subset of dict keys

    :param d: dict
    :param keys: list
    :return: dict
    """
    keys = set(keys).intersection(d.keys())

    sd = {k:d[k] for k in keys}

    return sd


def build_search_kwargs(d, **kwargs):
    """
    Builds kwargs for api.search

    Only includes non None items in d

    :param d: dict
    :param kwargs:
    :return: dict

    """
    nd = d.copy()
    for k, v in kwargs.items():
        if v is not None:
            nd[k] = v

    return nd


def string_to_list(s):
    """
    Convert argument string (of potentially a list of values) to a list of strings

    :param s: str
    :return: list[str]

    """
    if s is not None:
        s = [c.strip() for c in s.split(',')]
    return s


def string_to_epoch(s):
    """
    Convert argument string to epoch if possible

    If argument looks like int + s,h,md (ie, 30d), we'll pass as-is
    since pushshift can accept this.  Per docs, pushshift supports:
        Epoch value or Integer + "s,m,h,d" (i.e. 30d for 30 days)

    :param s: str
    :return: int | str

    """
    if s is not None:
        s = s.strip()
        if re.search('^[0-9]+[smhd]$', s):
            return s

        try:
            s = dp.parse(s).timestamp()
            s = int(s)
        except ValueError:
            raise click.BadParameter("could not convert argument to "
                                     "a datetime: {}".format(s))

    return s


class DummyProgressBar(object):
    """
    Dummy progress bar that just returns generator but displays no output.

    """
    def __init__(self, things):
        self.things = things

    def __enter__(self):
        return self.things

    def __exit__(self, type, value, traceback):
        pass



from setuptools import setup
from os import path
import re

PACKAGE_NAME='psaw'
HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.rst'), encoding='utf-8') as fp:
    README = fp.read()

with open(path.join(HERE, PACKAGE_NAME, '__init__.py'),
          encoding='utf-8') as fp:
    VERSION = re.search("__version__ = '([^']+)'", fp.read()).group(1)

setup(name=PACKAGE_NAME,
      packages=[PACKAGE_NAME],
      version=VERSION,
      long_description=README,
      description='Pushshift.io API Wrapper for reddit.com public comment/submission search',
      author='David Marx',
      author_email='david.marx84@gmail.com',
      url='http://github.com/dmarx/psaw',
      license='Simplified BSD License',
      install_requires=['requests', 'Click'],
      entry_points="""
          [console_scripts]
          psaw=psaw.psaw:cli
      """,
      keywords='reddit api wrapper pushshift',
      python_requires='>=3',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Utilities']
      )
